import hashlib
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from repoproviders.fetchers import fetch
from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import DoesNotExist


@pytest.mark.parametrize(
    ("questions", "md5tree"),
    [
        (
            # Fetch an immutable subdir set up in yuvipanda's drive and is public
            # We pick a small but still nested dir to make the test faster
            (
                "https://drive.google.com/drive/folders/12zoh_LubWyRMZG-KR9GGyukBsO8SCU9j",
            ),
            {
                "PULL_REQUEST_TEMPLATE.md": "0dc1f0612f21713189ed4ccc0d00197b",
                "SUPPORT.md": "14ddad47a068e1660d37fcef01026763",
                "workflows/release.yml": "0646032f1588e8b8621cd1c153bb6fbf",
                "workflows/test.yml": "d778015caa14985a6f61757de44bd5e0",
                "workflows/cron.yml": "3d055970af426e2065403d11dfa526b6",
                "workflows/main.yml": "70271a5c0e3dab1a25df2f6b3f8449eb",
                "ISSUE_TEMPLATE/02-question.yml": "7d7fceedd5b01853f575e0dff757f536",
                "ISSUE_TEMPLATE/03-feature-request.yml": "a1d4f142c8503aa7220220877416106a",
                "ISSUE_TEMPLATE/config.yml": "0311551ec2b71e6f8892ba688ea57dd7",
                "ISSUE_TEMPLATE/01-bug-report.yml": "c23bc48db00e2a5f48185f395614fbf7",
            },
        ),
    ],
)
async def test_fetch(questions: list[str], md5tree: dict[str, str]):
    for question in questions:
        with TemporaryDirectory() as d:
            output_dir = Path(d)
            answers = await resolve(question, True)

            assert answers is not None
            assert not isinstance(answers[-1], DoesNotExist)

            await fetch(answers[-1].repo, output_dir)

            # Verify md5 sum of the files we expect to find
            for subpath, expected_sha in md5tree.items():
                with open(output_dir / subpath, "rb") as f:
                    h = hashlib.md5()
                    h.update(f.read())
                    assert h.hexdigest() == expected_sha



import json
from dataclasses import dataclass
from shutil import which
from tempfile import NamedTemporaryFile

from repoproviders.resolvers.base import DoesNotExist, Exists

from ..utils import GCP_PUBLIC_SERVICE_ACCOUNT_KEY, exec_process, make_dir_hash


@dataclass(frozen=True)
class GoogleDriveFolder:
    id: str

    immutable = False


@dataclass(frozen=True)
class ImmutableGoogleDriveFolder:
    id: str
    dir_hash: str

    immutable = True


class GoogleDriveFolderResolver:
    async def resolve(
        self, question: GoogleDriveFolder
    ) -> Exists[ImmutableGoogleDriveFolder] | DoesNotExist[GoogleDriveFolder] | None:

        if not which("rclone"):
            raise FileNotFoundError(
                "rclone must be installed to resolve folders from Google Drive"
            )
        with NamedTemporaryFile("w") as service_account_key:
            json.dump(GCP_PUBLIC_SERVICE_ACCOUNT_KEY, service_account_key)
            service_account_key.flush()

            connection_string = f":drive,scope=drive.readonly,service_account_file={service_account_key.name}:"

            cmd = [
                "rclone",
                "lsjson",
                connection_string,
                "--recursive",
                "--hash",
                "--drive-root-folder-id",
                question.id,
            ]

            returncode, stdout, stderr = await exec_process(cmd)

            if returncode != 0:
                # Failure in one way or another. Let's just write out the failure message lines
                # that refer to lsjson, so we avoid messages about missing config files
                # Cut off first 20 chars, as it prints out the date
                stderr_lines = [l[20:] for l in stderr.splitlines()]
                message = " ".join(l for l in stderr_lines if "lsjson" in l)
                # FIXME: Does this leak sensitive info?
                return DoesNotExist(GoogleDriveFolder, message)

            data = json.loads(stdout)

            if len(data) == 0:
                # No items were returned. Let's treat this as a DoesNotExist, as this usually means we don't
                # have permissions, or the directory doesn't exist
                return DoesNotExist(
                    GoogleDriveFolder,
                    "The Google Drive Folder either does not exist, is empty or is not public",
                )

            hash_input = {}
            for item in data:
                # Use (in order of preference), sha256, sha1, md5 and modtime based on what is present
                hashes = item.get("Hashes", {})
                h = hashes.get(
                    "sha256", hashes.get("sha1", hashes.get("md5", item["ModTime"]))
                )
                hash_input[item["Path"]] = h

            dirhash = make_dir_hash(hash_input)

            return Exists(ImmutableGoogleDriveFolder(question.id, dirhash))



import json
import subprocess
from pathlib import Path
from shutil import which
from tempfile import NamedTemporaryFile

from repoproviders.resolvers.rclone import GoogleDriveFolder, ImmutableGoogleDriveFolder

from ..utils import GCP_PUBLIC_SERVICE_ACCOUNT_KEY


class GoogleDriveFolderFetcher:
    async def fetch(
        self, repo: ImmutableGoogleDriveFolder | GoogleDriveFolder, output_dir: Path
    ):
        if not which("rclone"):
            raise FileNotFoundError(
                "rclone must be installed to fetch folders from Google Drive"
            )
        # FIXME: We don't actually check the dirhash of the ImmutableGoogleDriveFolder, so it may have
        # mutated since we asked for things to be done to it. I don't exactly know what to do about that.
        with NamedTemporaryFile("w") as service_account_key:
            json.dump(GCP_PUBLIC_SERVICE_ACCOUNT_KEY, service_account_key)
            service_account_key.flush()

            connection_string = f":drive,scope=drive.readonly,service_account_file={service_account_key.name}:"
            rclone_cmd = [
                "rclone",
                "copy",
                connection_string,
                "--drive-root-folder-id",
                repo.id,
                str(output_dir),
            ]

            subprocess.check_call(rclone_cmd)



import hashlib
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from repoproviders.fetchers import fetch
from repoproviders.resolvers import resolve
from repoproviders.resolvers.base import DoesNotExist


@pytest.mark.parametrize(
    ("questions", "md5tree"),
    [
        (
            # These expire after 90 days. We should set up our own github automation for this.
            # However, instead, I'll buy us 90 days by using a recent run.
            # We expect this to fail roughly 90 days from now.
            # When that happens, you can either:
            # 1. Fix this properly by rewriting the test here to dynamically look for artifacts
            # 2. Go to https://github.com/jupyterlab/jupyterlab/actions/workflows/galata.yml, pick a
            #    recent completed run, find the URL for `documentation-test-assets` and update the content
            # https://github.com/yuvipanda/repoproviders/issues/28 has more information
            (
                "https://github.com/jupyterlab/jupyterlab/actions/runs/21958742242/artifacts/5487665511",
            ),
            {
                "test-documentation-workspa-aae0b-bar-Workspaces-context-menu-documentation/video.webm": "e2f7acd5785ed28ec6177081e8a7f7e8",
                "test-documentation-workspa-aae0b-bar-Workspaces-context-menu-documentation/error-context.md": "de239c63a0b17f6f3309e5afd2c47428",
                "test-documentation-customi-38159-t-should-use-default-layout-documentation/video.webm": "4f190c86be532f2d044099011306a38e",
                "test-documentation-customi-38159-t-should-use-default-layout-documentation/default-terminal-position-single-actual.png": "5a0a2040f0fc5b1457986f3575117433",
                "test-documentation-customi-38159-t-should-use-default-layout-documentation/default-terminal-position-single-diff.png": "a042a221e77dcc4954a0fa4e912eee9b",
                "test-documentation-customi-38159-t-should-use-default-layout-documentation/error-context.md": "8e72e868618f8ac3052df7ccc1689e5a",
                "test-documentation-customi-38159-t-should-use-default-layout-documentation/default-terminal-position-single-expected.png": "0ef5ee2bada38346cc2df97b1e4d16f1",
            },
        ),
    ],
)
async def test_fetch(questions: list[str], md5tree: dict[str, str]):
    for question in questions:
        with TemporaryDirectory() as d:
            output_dir = Path(d)
            answers = await resolve(question, True)

            assert answers is not None
            assert not isinstance(answers[-1], DoesNotExist)

            await fetch(answers[-1].repo, output_dir)

            # Verify md5 sum of the files we expect to find
            for subpath, expected_sha in md5tree.items():
                with open(output_dir / subpath, "rb") as f:
                    h = hashlib.md5()
                    h.update(f.read())
                    assert h.hexdigest() == expected_sha



import pytest

from repoproviders.resolvers.base import DoesNotExist, Exists
from repoproviders.resolvers.rclone import (
    GoogleDriveFolder,
    GoogleDriveFolderResolver,
    ImmutableGoogleDriveFolder,
)


@pytest.mark.parametrize(
    ("url", "expected"),
    (
        # Immutable directory
        (
            GoogleDriveFolder(id="1OBwu72mlrWymv8DLepOwPY-GWHPrgYN8"),
            Exists(
                ImmutableGoogleDriveFolder(
                    id="1OBwu72mlrWymv8DLepOwPY-GWHPrgYN8",
                    dir_hash="hHea2vn_tW34FEXigXp2Eurs3adDso6XYgdeRhzPWPo=",
                )
            ),
        ),
        # Empty but public directory
        (
            GoogleDriveFolder("1sokbMclA4UaiXPdBBGyThlBEGv961psp"),
            DoesNotExist(
                GoogleDriveFolder,
                "The Google Drive Folder either does not exist, is empty or is not public",
            ),
        ),
        # Invalid ID
        (
            GoogleDriveFolder(id="1OBwu72mlrWymv8DLepOwPY-GWHPr8"),
            DoesNotExist(
                GoogleDriveFolder,
                "NOTICE: Failed to lsjson: error in ListJSON: couldn't list directory: googleapi: Error 404: File not found: ., notFound",
            ),
        ),
        # Private unshared folder
        (
            GoogleDriveFolder(id="1lvm0Co_aYa0iC__5QLyeL2ptcCNFpsIL"),
            DoesNotExist(
                GoogleDriveFolder,
                "The Google Drive Folder either does not exist, is empty or is not public",
            ),
        ),
    ),
)
async def test_gist(url, expected):
    gh = GoogleDriveFolderResolver()
    assert await gh.resolve(url) == expected



from pathlib import Path
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

import aiohttp
from yarl import URL

from repoproviders.resolvers.repos import GitHubActionArtifact

from ..resolvers.git import ImmutableGit
from ..utils import GITHUB_PUBLIC_PAT, download_file, exec_process


class ImmutableGitFetcher:
    async def fetch(self, repo: ImmutableGit, output_dir: Path):

        # Assume output_dir is empty
        command = [
            "git",
            "clone",
            "--filter=tree:0",
            "--recurse-submodules",
            repo.repo,
            str(output_dir),
        ]
        retcode, stdout, stderr = await exec_process(command)

        if retcode != 0:
            # FIXME: Raise a more helpful error?
            raise CalledProcessError(retcode, command, stdout, stderr)

        command = ["git", "checkout", repo.ref]

        retcode, stdout, stderr = await exec_process(command, cwd=str(output_dir))

        if retcode != 0:
            # FIXME: Raise a more helpful error?
            raise CalledProcessError(retcode, command, stdout, stderr)


class GitHubActionArtifactFetcher:
    async def fetch(self, repo: GitHubActionArtifact, output_dir: Path):
        # Assume this standard archive URL holds
        # If this starts to fail, we shall make an additional GitHub API request
        # to this URL: https://docs.github.com/en/rest/actions/artifacts?apiVersion=2022-11-28#get-an-artifact
        # FIXME: Support other installations
        download_url = (
            URL("https://api.github.com/repos/")
            / repo.account
            / repo.repo
            / "actions/artifacts"
            / str(repo.artifact_id)
            / "zip"
        )

        async with aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {GITHUB_PUBLIC_PAT}"}
        ) as session:
            with NamedTemporaryFile() as temp_file:
                await download_file(session, download_url, Path(temp_file.name))
                compressed_file = ZipFile(temp_file.name)
                compressed_file.extractall(output_dir)



import asyncio
import binascii
import hashlib
import json
from base64 import standard_b64decode, urlsafe_b64encode
from pathlib import Path

import aiohttp
from yarl import URL

# A public service account JSON key, used to make API requests for Google Drive.
# Some values are base 64 decoded so we don't get dinged by GitHub security scanning
# This service account has no rights, only used for identity
GCP_PUBLIC_SERVICE_ACCOUNT_KEY = {
    "type": "service_account",
    "project_id": "repoproviders-iam",
    "private_key_id": standard_b64decode(
        "MDY5ODJhNmY4ZWM5MTM3MDU0YmU5Mjk5NzIwYTQ1OWFhYjgxMmY0Yg=="
    ).decode(),
    "private_key": standard_b64decode(
        "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"
    ).decode(),
    "client_email": "repoproviders-no-rights-public@repoproviders-iam.iam.gserviceaccount.com",
    "client_id": "107622683369583114795",
    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
    "token_uri": "https://oauth2.googleapis.com/token",
    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
    "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/repoproviders-no-rights-public%40repoproviders-iam.iam.gserviceaccount.com",
    "universe_domain": "googleapis.com",
}

# Publicly visible GitHub PAT with just public repo permissions of account yuvipanda-repoproviders-public
# Primarily useful for downloading GitHubActionsArtifact resources, as you need a PAT to do so
GITHUB_PUBLIC_PAT = binascii.a2b_uu(
    b"H9VAP7UE%9&%N;U5O34YF4712:%HQ9VQP,VUF,U),3D%J9S%Q8GHW:P  \n"
).decode()


def make_dir_hash(data: dict[str, str]) -> str:
    """
    Convert a directory structure to a urlsafe base64 encoded sha256 hash.

    Input is a dict where key is full paths of files, and value is either a hash or a last modified
    timestamp.  Sorted by key explicitly before hashing.
    """
    data = dict(sorted(data.items()))
    return urlsafe_b64encode(
        hashlib.sha256(json.dumps(data).encode()).digest()
    ).decode()


async def download_file(session: aiohttp.ClientSession, url: URL, output_path: Path):
    # Read in 4k chunks
    CHUNK_SIZE = 4 * 1024
    resp = await session.get(url)

    if resp.status == 200 and "Location" in resp.headers:
        # Some providers (lookin at you, data.caltech.edu) send a Location header
        # *but with a 200 status code*. This is invalid and bogus, yet we have to
        # honor it. Sigh
        return await download_file(session, URL(resp.headers["Location"]), output_path)

    resp.raise_for_status()

    if not output_path.parent.exists():
        # Leave the exist_ok=True here despite the check, to handle possible
        # race conditions in the future if we ever parallelize this
        output_path.parent.mkdir(parents=True, exist_ok=True)

    with open(output_path, "wb") as f:
        async for chunk in resp.content.iter_chunked(CHUNK_SIZE):
            f.write(chunk)


async def exec_process(cmd: list[str], **kwargs) -> tuple[int, str, str]:
    """
    Execute given command and return return code, stdout, stderr
    """

    proc = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, **kwargs
    )
    stdout, stderr = await proc.communicate()
    returncode = await proc.wait()

    return returncode, stdout.decode(), stderr.decode()
